import 'dart:io';
import 'dart:async';
import 'package:adeptarch_base/adeptarch_base.dart';
import 'package:adeptgraphql/adeptgraphql.dart';
import 'package:adeptutils/adeptutils.dart' as AdeptUtils;
import 'package:flutter/foundation.dart';
import 'package:graphql_flutter/graphql_flutter.dart';
import 'package:graphql/client.dart';
import 'package:http/http.dart' as http;
import 'package:web_socket_channel/web_socket_channel.dart';

GraphQLAPI graphQLAPI = GraphQLAPI();

/// This class works as abstraction for GraphQL api connections

class GraphQLAPI {
  late AdeptGraphQLClient client;
  ValueNotifier<AdeptGraphQLClient> clientNotifier;
  http.Client? httpClient;
  GraphQLCache? cache;
  late Link apiLink;
  String apiUrl;
  String? wsApiUrl;
  Duration? timeout;

  GraphQLAPI()
      : clientNotifier = ValueNotifier<AdeptGraphQLClient>(AdeptGraphQLClient(
          link: HttpLink(''),
          cache: GraphQLCache(),
        )),
        apiUrl = '';

  void init(
      {Link? link,
      String? apiUrl,
      String? wsApiUrl,
      GetToken? getToken,
      Duration? timeout,
      SocketClientConfig? socketClientConfig,
      void Function(dynamic event)? onSocketData,
      void Function()? onSocketConnected,
      void Function()? onSocketDisconnected,
      void Function(dynamic error)? onSocketError,
      http.Client? httpClient}) {
    if (apiUrl != null) this.apiUrl = apiUrl;
    if (wsApiUrl != null) this.wsApiUrl = wsApiUrl;
    if (timeout != null) this.timeout = timeout;
    if (httpClient != null) this.httpClient = httpClient;

    if (link != null) {
      apiLink = link;
    } else {
      AuthLink? authLink =
          getToken != null ? AuthLink(getToken: getToken) : null;
      HttpLink httpLink = HttpLink(this.apiUrl, httpClient: this.httpClient);
      WebSocketLink? webSocketLink;
      if (this.wsApiUrl?.isNotEmpty ?? false) {
        if (onSocketData != null ||
            onSocketConnected != null ||
            onSocketDisconnected != null ||
            onSocketError != null) {
          socketClientConfig ??= const SocketClientConfig();
          socketClientConfig = SocketClientConfig(
              autoReconnect: socketClientConfig.autoReconnect,
              inactivityTimeout: socketClientConfig.inactivityTimeout,
              queryAndMutationTimeout:
                  socketClientConfig.queryAndMutationTimeout,
              delayBetweenReconnectionAttempts:
                  socketClientConfig.delayBetweenReconnectionAttempts,
              headers: socketClientConfig.headers,
              initialPayload: socketClientConfig.initialPayload,
              parser: socketClientConfig.parser,
              serializer: socketClientConfig.serializer,
              connectFn: (uri, protocols) {
                var channel =
                    WebSocketChannel.connect(uri, protocols: protocols);
                // without this line, our client won't be able to listen to stream events,
                // because you are already listening.
                channel = channel.forGraphQL();
                channel.stream.listen((event) {
                  // Socket data received
                  onSocketData?.call(event);
                  if (event == '{"type":"connection_ack"}' ||
                      (event is Map && event['type'] == 'connection_ack')) {
                    // Socket connected
                    onSocketConnected?.call();
                  }
                }, onError: (error) {
                  onSocketError?.call(error);
                }, onDone: () {
                  onSocketDisconnected?.call();
                });
                return channel;
              });
        }
        webSocketLink = socketClientConfig != null
            ? WebSocketLink(
                this.wsApiUrl!,
                config: socketClientConfig,
              )
            : WebSocketLink(
                this.wsApiUrl!,
              );
      }

      if (authLink != null) {
        apiLink = authLink;
        apiLink = apiLink.concat(httpLink);
      } else {
        apiLink = httpLink;
      }
      if (webSocketLink != null) {
        apiLink = Link.split(
            (request) => request.isSubscription, webSocketLink, apiLink);
      }
    }

    client = AdeptGraphQLClient(
      link: apiLink,
      cache: cache ?? GraphQLCache(),
      timeout: timeout,
    );
    clientNotifier.value = client;
  }

  ///Implement this method when checked what exceptions could be
  String? getLocalizedMessageFromException(dynamic exception) {
    if (exception == null) return null;
    if (exception is AdeptUtils.GenericException) {
      return getLocalizedMessageFromException(exception.causeException);
    }

    if (exception is WebSocketException ||
        exception is TimeoutException ||
        exception.toString().toLowerCase().contains("TimeoutException")) {
      return "Server connection timeout.";
    }
    if (exception is SocketException ||
        exception.toString().toLowerCase().contains("failed to connect")) {
      return "Check your internet connection. ${exception.toString()}";
    }
    if (exception is HandshakeException) return exception.toString();

    return exception.toString();
  }

  String? getGenericLocalizedMessageFromException(dynamic exception) {
    String? message = getLocalizedMessageFromException(exception);
    return message ?? "Unknown error";
  }

  String? getLocalizedMessageFromHttpResultCode(int? code) {
    if (code == null) return null;
    switch (code) {
      case HttpStatus.noContent:
        return "($code) No Content";

      case HttpStatus.badRequest:
        return "($code) Bad Request";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized";

      case HttpStatus.forbidden:
        return "($code) Forbidden";

      case HttpStatus.notFound:
        return "($code) Not Found";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error";

      default:
        return "HTTP error code: ($code)";
    }
  }

  String? getLocalizedMessageFromResponse(
      {QueryResponse? response, int? code}) {
    code ??= response?.errorStatusCode;
    if (code == null) return null;
    String extraMessage = " ${response?.errorMessage ?? ""}";
    switch (code) {
      case HttpStatus.noContent:
        return "($code) No Content.$extraMessage";

      case HttpStatus.badRequest:
        return "($code) Bad Request.$extraMessage";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized.$extraMessage";

      case HttpStatus.forbidden:
        return "($code) Forbidden.$extraMessage";

      case HttpStatus.notFound:
        return "($code) Not Found.$extraMessage";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error.$extraMessage";

      case AdeptUtils.Jsonable.jsonParserError:
        return "($code) Invalid response. JsonParser was unable to decode response.$extraMessage";
    }
    return "HTTP error code: ($code).$extraMessage";
  }
}
