import 'package:adeptarch_base/adeptarch_base.dart';
import 'package:adeptgraphql/src/api/query_response.dart';

/// A generic class that can provide a resource backed by both the local database and the network.

typedef LoadFromGraphQLAPI<DataType, ErrorType>
    = Future<QueryResponse<DataType, ErrorType>> Function();
typedef OnLoadedFromGraphQLAPISuccess<DataType, ErrorType>
    = Future<Resource<DataType>> Function(
        {Resource<DataType>? resource,
        QueryResponse<DataType, ErrorType>? response});
typedef OnLoadedFromGraphQLAPIFailed<DataType, ErrorType>
    = Future<Resource<DataType>> Function(
        {dynamic exception, QueryResponse<DataType, ErrorType>? response});

typedef OnGraphQLAPIErrorManagement<DataType, ErrorType>
    = Future<Resource<DataType>> Function(Resource<DataType>? resource,
        {QueryResponse<DataType, ErrorType>? response});

class QueryResourceLoader<DataType, ErrorType>
    extends ResourceLoader<DataType> {
  LoadFromGraphQLAPI<DataType?, ErrorType?>? loadFromGraphQLAPI;
  OnLoadedFromGraphQLAPISuccess<DataType?, ErrorType?>?
      onLoadedFromGraphQLAPISuccess;
  OnLoadedFromGraphQLAPIFailed<DataType?, ErrorType?>?
      onLoadedFromGraphQLAPIFailed;

  OnGraphQLAPIErrorManagement<DataType?, ErrorType?>?
      onGraphQLAPIErrorManagement;

  QueryResourceLoader({
    AdeptPublishSubject<Resource<DataType?>>? publishSubject,
    LoadFromCache<DataType?>? loadFromCache,
    OnLoadedFromCacheSuccess<DataType?>? onLoadedFromCacheSuccess,
    OnLoadedFromCacheFailed<DataType?>? onLoadedFromCacheFailed,
    LoadFromStorage<DataType?>? loadFromStorage,
    OnLoadedFromStorageSuccess<DataType?>? onLoadedFromStorageSuccess,
    OnLoadedFromStorageFailed<DataType?>? onLoadedFromStorageFailed,
    this.loadFromGraphQLAPI,
    this.onLoadedFromGraphQLAPISuccess,
    this.onLoadedFromGraphQLAPIFailed,
    OnFailed<DataType?>? onFailed,
    this.onGraphQLAPIErrorManagement,
    bool stopOnFirstResourceLoaded = false,
  }) : super(
          publishSubject: publishSubject,
          loadFromCache: loadFromCache,
          onLoadedFromCacheSuccess: onLoadedFromCacheSuccess,
          onLoadedFromCacheFailed: onLoadedFromCacheFailed,
          loadFromStorage: loadFromStorage,
          onLoadedFromStorageSuccess: onLoadedFromStorageSuccess,
          onLoadedFromStorageFailed: onLoadedFromStorageFailed,
          onFailed: onFailed,
          stopOnFirstResourceLoaded: stopOnFirstResourceLoaded,
        );

  @override
  bool get loadMoreAfterCacheLoaded =>
      !stopOnFirstResourceLoaded &&
      (loadFromStorage != null || loadFromGraphQLAPI != null);

  @override
  bool get loadMoreAfterStorageLoaded =>
      !stopOnFirstResourceLoaded && loadFromGraphQLAPI != null;

  ///GraphQL API handlers
  Future<Resource<DataType?>> handleLoadFromGraphQLAPI() async {
    try {
      if (loadFromGraphQLAPI == null) {
        return Resource(status: ResourceStatus.TRANSIENT);
      }
      final QueryResponse<DataType?, ErrorType?>? response =
          await loadFromGraphQLAPI?.call();
      if (response?.isSuccessful ?? false) {
        return handleOnLoadedFromGraphQLAPISuccess(response: response);
      } else {
        return handleOnLoadedFromGraphQLAPIFailed(response: response);
      }
    } catch (e) {
      return handleOnLoadedFromGraphQLAPIFailed(exception: e);
    }
  }

  Future<Resource<DataType?>> handleOnLoadedFromGraphQLAPISuccess(
      {Resource<DataType?>? resource,
      QueryResponse<DataType?, ErrorType?>? response}) async {
    if (onLoadedFromGraphQLAPISuccess != null) {
      return onLoadedFromGraphQLAPISuccess!(
          resource: resource, response: response);
    }
    if (response?.isSuccessful ?? false) {
      return handleOnLoadedFromGraphQLAPISuccess(
          resource: Resource<DataType>.success(response?.data,
              extras: response?.source));
    }
    resource ??= Resource<DataType>.error(null, 'Empty GraphQL API response');
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  Future<Resource<DataType?>> handleOnLoadedFromGraphQLAPIFailed(
      {dynamic exception,
      QueryResponse<DataType?, ErrorType?>? response}) async {
    if (onLoadedFromGraphQLAPIFailed != null) {
      return onLoadedFromGraphQLAPIFailed!(
          exception: exception, response: response);
    }
    DataType? data = response?.data ??
        (publishSubjectHasValue ? publishSubject?.value?.data : null);
    exception ??= response?.exception;
    Resource<DataType?> resource = Resource<DataType>.error(
        data, response?.errorMessage,
        code: response?.errorStatusCode, exception: exception);
    if (onGraphQLAPIErrorManagement != null) {
      resource =
          await onGraphQLAPIErrorManagement!(resource, response: response);
    }
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  ///
  /// Loads a resource from cache -> storage -> GraphQL API.
  /// User can opt in to stopOnFirstResourceLoaded, which means when a resource
  /// is completed loaded (dataLoaded != null) then no further load will be made
  ///
  @override
  Future<Resource<DataType?>> loadResource() async {
    Resource<DataType?> resource =
        Resource<DataType?>(status: ResourceStatus.TRANSIENT);
    try {
      if (loadFromCache != null) resource = await handleLoadFromCache();
      if (hasResourceCompletedLoading(resource)) return resource;
      if (loadFromStorage != null) resource = await handleLoadFromStorage();
      if (hasResourceCompletedLoading(resource)) return resource;
      if (loadFromGraphQLAPI != null) {
        resource = await handleLoadFromGraphQLAPI();
      }
      if (!hasResourceCompletedLoading(resource)) {
        resource.status = ResourceStatus.SUCCESS;
        if (publishResults) publishSubject?.post(resource);
      }
    } catch (e) {
      resource = await handleOnFailed(e);
    }
    return resource;
  }
}
